/* Copyright (C) Kevin Vance, June 26 2003.
 * Email winnocence@peacefire.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>

#include "hostsfile.h"

/* States while reading from hosts file */
enum { SEARCHING = 0, COPYING = 1, FINISHED = 2 };

/* Markers and default block list */
static char begin_marker[] = "###BEGIN LINES INSERTED BY WINNOCENCE\r\n";
static char defaults[] = "www.playboy.com\r\n" \
			 "www.hustler.com\r\n" \
			 "www.sex.com\r\n";
static char end_marker[]   = "###END LINES INSERTED BY WINNOCENCE\r\n";

/* append(original, size, new)
 * ---------------------------
 * Append the contents of the new buffer into the original, resizing if
 * necessary. */
void append(char **original, unsigned int *size, char *new)
{
	if(strlen(*original)+strlen(new)+1 > *size) {
		*size += strlen(new);
		*original = realloc(*original, *size);
	}
	strcat(*original, new);
}

/* fgetline(fp, buffer, size)
 * --------------------------
 * Read one line from the file, growing the buffer if necessary */
int  _fg_len = -1;
int  _fg_pos = -1;
int  _fg_i   = 0;
char _fg_buf[1024];
void fgetline(FILE *fp, char **buffer, unsigned int *size)
{
	_fg_i = 0;

	while(1) {
		if(_fg_pos >= _fg_len) {
			/* Reading 1k at a time is faster than 1 byte */
			_fg_len = fread(_fg_buf, 1, 1024, fp);
			_fg_pos = 0;
		}
		if(_fg_len <= 0) {
			_fg_len = _fg_pos = -1;
			break;
		}
		if(_fg_i + _fg_len + 1 == *size) {
			/* Add another 1k if the line is too long */
			*size += 1024;
			*buffer = realloc(*buffer, *size);
		}
		for(; _fg_pos < _fg_len; _fg_pos++, _fg_i++) {
			(*buffer)[_fg_i] = _fg_buf[_fg_pos];
			if(_fg_buf[_fg_pos] == '\n') {
				/* End of line, finished! */
				_fg_pos++;
				_fg_i++;
				goto finished;
			}
		}
	}
finished:
	(*buffer)[_fg_i] = '\0';
}


/* get_hostsfile()
 * ---------------
 * Find the location of the hosts file based on the OS version */
char *get_hostsfile()
{
	char *windir = getenv("WINDIR");
	static char suffix_95[] = "\\hosts";
	static char suffix_NT[] = "\\system32\\drivers\\etc\\hosts";
	OSVERSIONINFOEX winver;
	char *retval;

	if(windir == NULL)
		/* If this isn't right, MAYBE YOU SHOULD SET YOUR ENVIRONMENT
		 * CORRECTLY! */
		windir = "C:\\windows";

	winver.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
	GetVersionEx((LPOSVERSIONINFO)&winver);
	if(winver.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS) {
		/* Windows 95, 98, or ME */
		retval = malloc(strlen(windir)+strlen(suffix_95)+1);
		strcpy(retval, windir);
		strcat(retval, suffix_95);
	} else {
		/* Windows NT, 2K, or XP */
		retval = malloc(strlen(windir)+strlen(suffix_NT)+1);
		strcpy(retval, windir);
		strcat(retval, suffix_NT);
	}

	return retval;
}

/* get_winnocence_lines()
 * ----------------------
 * Load the winnocence lines from the system hosts file */
int _hf_defaults = 0;
char *get_winnocence_lines()
{
	int state = SEARCHING;
	/* Current line */
	unsigned int linesize = 2048;
	char *line = malloc(linesize);
	/* Returned lines */
	unsigned int retsize = 2048;
	char *retval = malloc(retsize);
	/* Load hosts file */
	char *fname = get_hostsfile();
	FILE *fp = fopen(fname, "a+b");
	free(fname);
	_hf_defaults = 0;

	if(fp == NULL) {
		free(line);
		free(retval);
		return NULL;
	}
	
	retval[0] = '\0';
	fseek(fp, 0, SEEK_SET);
	while(state != FINISHED) {
		fgetline(fp, &line, &linesize);
		if(line[0] == '\0')
			break;
		if(state == SEARCHING && !strcmp(line, begin_marker))
			state = COPYING;
		else if(state == COPYING && !strcmp(line, end_marker))
			state = FINISHED;
		else if(state == COPYING) {
			if(sscanf(line, "209.211.253.169 %[^\b]", line))
				append(&retval, &retsize, line);
		}
	}

	fclose(fp);
	free(line);

	if(state != FINISHED) {
		/* We didn't find our lines.  Use defaults. */
		_hf_defaults = 1;
		strcpy(retval, defaults);
	}

	return retval;
}

/* put_winnocence_lines(blocked_sites)
 * -----------------------------------
 * Write our lines to the hosts file.  Return 0 for error, non-0 for OK. */
int put_winnocence_lines(char *blocked_sites)
{
	int state = SEARCHING;
	/* Current line */
	unsigned int linesize = 2048;
	char *line = malloc(linesize);

	char temppath[MAX_PATH];
	char tempfname[MAX_PATH];
	char *hostsloc;
	FILE *out, *in;

	/* Create temp file */
	GetTempPath(MAX_PATH, temppath);
	GetTempFileName(temppath, "win", 0, tempfname);
	out = fopen(tempfname, "wb");
	if(out == NULL) {
		free(line);
		return 0;
	}

	/* Load current file */
	hostsloc = get_hostsfile();
	in = fopen(hostsloc, "rb");
	if(in == NULL) {
		free(line);
		free(hostsloc);
		return 0;
	}

	while(1) {
		fgetline(in, &line, &linesize);
		if(line[0] == '\0') {
			if(state != FINISHED) {
				/* First time writing block list. */
				fwrite("\r\n", 1, 2, out);
				write_lines(out, blocked_sites);
			}
			break;
		}

		if(state == SEARCHING && !strcmp(line, begin_marker)) {
			/* Found begin marker--write block list. */
			state = COPYING;
			write_lines(out, blocked_sites);
		} else if(state == COPYING && !strcmp(line, end_marker)) {
			/* Found end marker--resume copying lines */
			state = FINISHED;
		} else if(state == COPYING)
			/* Inside old block list, ignore. */;
		else
			/* Before/After end marker--keep copying. */
			fwrite(line, 1, _fg_i, out);
	}

	fclose(out);
	fclose(in);

	{
	int retval;
	if(CopyFile(tempfname, hostsloc, FALSE) == FALSE)
		retval = 0;
	else
		retval = 1;

	DeleteFile(tempfname);

	free(line);
	free(hostsloc);

	return retval;
	}
}

/* write_lines(fp, blocked_sites)
 * ------------------------------
 * Write our site list to the given file, in hosts format. */
int write_lines(FILE *fp, char *blocked_sites)
{
	char *token = strtok(blocked_sites, "\n");
	fputs(begin_marker, fp);
	while(token != NULL) {
		fprintf(fp, "209.211.253.169 %s\n", token);
		token = strtok(NULL, "\n");
	}
	fputs(end_marker, fp);
	return 0;
}
